(function($){
    /**
     * The API boundary is:
     * - The TD/TH belongs to spreadsheet-macro.js,
     * - The renderer is free with the innerHTML,
     * - The CSS class of the td/th is part of the API. The spreadsheet communicates with the renderers
     * through the methods of RendererBase; the renderers communicate with the spreadsheet using
     * the object passed in forCell(spreadsheet, $cell).
     *
     * Cells have a state cycle represented by the CSS classes:
     * - none: The cell is in view mode.
     * - editing: The cell contains an HTML input or any kind of data-entry element.
     * cleaned by forCell().edit() as the last resort.
     * - saving: The cell is being saved. We're waiting for server feedback and we don't
     * know yet whether it is saved. Once we receive it we'll immediately switch to
     * view or invalid mode.
     * - invalid: The value in the cell is invalid. It can't be saved.
     *
     * Other css classes:
     * - selected, focused
     * - formula: The cell contains a formula / the formula editor. The 'formula' class
     * is supposed to be there as soon as we know it's a formula (=when we write the TD)
     * or when we detect it's a formula (=when the user types '=').
     * - anti-formula: The cell starts with a '=' but is not a formula. This class is set
     * at the same time as 'formula' (when we build the TD or when the user clicks 'remove
     * formula') and will be removed if the user removes the '='.
     */
    AJS.bind("playsql-spreadsheet-init", function(event, api) {

        function outerHtml($value) {
            return $(document.createElement("div")).html($value).html();
        }

        var RendererBase = function() {
            return {
                key : "base-renderer",
                getI18nName : function() {
                    // Default implementation, please override...
                    return this.key;
                },
                isSystemRenderer : function() {
                    return true;
                },
                isSupportedType : function() {
                    throw "Not Implemented";
                },
                getDefaultType : function() {
                    throw "Not Implemented";
                },
                renderCellContents: function(value, spreadsheet, column) {
                    if (value === null || value === undefined) {
                        return "<span class='not-a-value'>&nbsp;</span>";
                    }
                    value = AJS.escapeHtml(value);
                    if (column.formats.css) {
                        var styles = $.parseJSON(column.formats.css),
                            styleAttr = "";
                        for (var key in styles) {
                            styleAttr = key + ":" + styles[key] + ";";
                        }
                        value = outerHtml($("<span>", {
                            style: styleAttr,
                            html: value
                        })).replace(/\n/g, "<br>\n");
                    }

                    return value;
                },
                onColumnDisplay : function(spreadsheet, column) {
                    // This method is triggered whenever the table is done displaying.
                    // Place additional code about cell rendering: binding handlers,
                    // displaying pictures or mapping external data.
                },
                createInputBox: function(value) {
                    return $("<input>", {
                        type: 'text',
                        value: value
                    });
                },
                /** This is only a helper method you can use to override createInputBox */
                createTextareaBox: function(value) {

                    function resizeTextarea() {
                        var $this = $(this);
                        $(this).height(this.scrollHeight);
                    }

                    window.setTimeout(function(){
                        $(".playsql textarea.uninitialized, .playsql-form-macro textarea.uninitialized").each(function(){
                            $(this).removeClass("uninitialized");
                            resizeTextarea.apply(this);
                        });
                    }, 0);

                    return $("<textarea>", {
                        text: value,
                        height: 1,
                        "class": "uninitialized"
                    }).on("keyup.playsqlresizetextarea keydown.playsqlresizetextarea change.playsqlresizetextarea", resizeTextarea);
                },
                autoCompleteFilter: function(spreadsheet, column, term, data, callback){

                    if (term.length > 0) {
                        Confluence.PlaySQL.Spreadsheet.ColumnToolbar.autoCompleteFilterString(column, term, data);
                    }
                    Confluence.PlaySQL.Spreadsheet.ColumnToolbar.autoCompleteFilterSQL(column, term, data);
                    callback();
                },
                /**
                 * Prepares the cell for edition
                 * @param spreadsheet
                 * @param $cell
                 * @param setValueByRef an array of the value to set. If provided, it will return "formula" if it
                 * starts with "=" and "ok" otherwise. It means it will override the status of the cell being a formula.
                 * @return the status:
                 *  - "ok" if it can be edited,
                 *  - "set" if the cell is already being edited and only the value must be set
                 *  - "formula" if it must be edited as a formula,
                 *  - "no" if it can't be edited
                 */
                checkBeforeEdit : function(spreadsheet, $cell, setValueByRef) {
                    if (!spreadsheet.perms.hasPerm('EDIT_CELLS') && !spreadsheet.perms.hasPerm('Q_VIEW_EDIT', spreadsheet.colOf($cell))) {
                        return "no";
                    }
                    if (setValueByRef[0]) {
                        if (setValueByRef[0].charAt(0) == "=" && !$cell.hasClass("anti-formula")) {
                            if (spreadsheet.perms.hasPerm('EDIT_ROWS_METADATA')) {
                                return "formula";
                            } else {
                                $cell.addClass("anti-formula");
                            }
                        }
                    } else {
                        if ($cell.hasClass("formula")) {
                            if (spreadsheet.perms.hasPerm('EDIT_ROWS_METADATA')) {
                                setValueByRef[0] = spreadsheet.getCellMetadata($cell, "formula");
                                return "formula";
                            } else {
                                $cell.addClass("anti-formula").removeClass("formula");
                            }
                        }
                    }
                    if ($cell.hasClass("editing") || $cell.hasClass("saving")) {
                        return "set";
                    }
                    if (setValueByRef[0] == "'") {
                        $cell.addClass("anti-formula");
                        setValueByRef[0] = "";
                    }
                    $cell.removeClass("formula");
                    // Check if it's a joint table
                    var recordIds = spreadsheet.getRecordIds($cell, true);
                    if (recordIds.jointIdName && !recordIds.jointId) {
                        spreadsheet.addJointRow(recordIds.jointIdCell.$cell);
                    }

                    return "ok";
                },
                forCell: function(spreadsheet, $cell) {
                    var parent = this,
                        operations = {
                            parent : parent,
                            /**
                             * Switch the cell to edit mode
                             * @param setValue optional. Sets the value. If so, it will re-evaluate whether it's a formula. It means
                             * if you pass a value which doesn't start with '=', it will set this value, replacing the formula.
                             */
                            edit: function(setValue) {
                                var setValueByRef = [setValue],
                                    situation = Confluence.PlaySQL.Spreadsheet.rendererBase.checkBeforeEdit(spreadsheet, $cell, setValueByRef);
                                // If the value was changed, use it
                                setValue = setValueByRef[0];

                                switch (situation) {
                                    case "no": return;
                                    case "formula":
                                        /* TODO Formulas activation/deactivation here */
                                        Confluence.PlaySQL.Spreadsheet.FormulaEditor.edit(spreadsheet, $cell, this, setValue);
                                        return;
                                    case "ok": break;
                                    case "set": break;
                                        setValue = this.parseValue(setValue);
                                        if (setValue != undefined) {
                                            $cell.find("input, textarea").val(setValue);
                                            $cell.find("div[contenteditable]").text(setValue);
                                        }
                                        return;
                                    default:
                                        throw "Not Implemented: situation " + situation;
                                }

                                setValue = this.parseValue(setValue);

                                $cell.addClass("editing");
                                // Keep the old value for cancel
                                $cell.attr('data-rendered', $cell.html());
                                var $input = parent.createInputBox(setValue != undefined ? setValue : $cell.attr('data-value'));

                                $input.appendTo($cell.empty());

                                var rendererInstance = this;
                                $input.keyup(function(){
                                    rendererInstance.validate(true);
                                }).on("blur.playsql-autocommit", function(){
                                    var $this = $(this),
                                        $cell = $this.closest("td");
                                    if (!$cell.hasClass("aui-dropdown2-trigger") && !$cell.hasClass("no-autocommit")) {
                                        window.setTimeout(function() {
                                            rendererInstance.saveAndView();
                                            spreadsheet.sprayRequestPipe();
                                        }, 0);
                                    }
                                });
                            },
                            parseValue : function(value) {
                                if (value == undefined) return value;
                                return $.trim(value);
                            },
                            /**
                             * Sets the focus on whatever HTML widget is central to this cell.
                             *
                             * No effect if the cell isn't in 'editing'.
                             *
                             * Beware to call this method on a 'click', not on a 'mousedown',
                             * because the 'click' event after a 'mousedown' resets the focus to the
                             * clicked element.
                             */
                            focus: function(withCaret, selEnd) {
                                if ($cell.hasClass("editing") && !spreadsheet.disableFocusManagement) {
                                    var $input = checkNotNull($cell.find("input, textarea, div[contenteditable]")).focus();
                                    if (withCaret == undefined)
                                        $input.select();
                                    else if (withCaret == true) {
                                        var el = $input[0],
                                            value = el.value || el.textContent;
                                        el.selectionStart = el.selectionEnd = value.length;
                                    } else if (typeof withCaret == "number") {
                                        var el = $input[0],
                                            value = el.value || el.textContent;
                                        el.selectionStart = withCaret;
                                        el.selectionEnd = selEnd != undefined ? selEnd : value.length;
                                    }
                                }
                            },
                            /** Returns from any state to view, by rendering the value which was provided by the server side */
                            view: function(value) {
                                // Allow for the value being null
                                if (value === undefined) {
                                    value = $cell.attr("data-value");
                                }
                                var viewRenderer = spreadsheet.getCell($cell).viewRenderer;
                                var rendered = (viewRenderer || this).parent.renderCellContents(value, spreadsheet, spreadsheet.colOf($cell));
                                $cell.html(rendered);
                                $cell.removeClass("editing saving invalid");
                                $cell.tipsy("disable");
                                $(".tipsy").remove();
                            },
                            getValue: function() {
                                return $cell.attr("data-value");
                            },
                            setOriginalValue: function(formulaValue, backendValue) {
                                this.view(formulaValue);
                                if (backendValue !== undefined) {
                                    $cell.attr("data-value", backendValue);
                                } else {
                                    $cell.attr("data-value", formulaValue);
                                }
                            },
                            validate: function(transformIfFormula) {
                                if ($cell.hasClass("formula")) {
                                    return Confluence.PlaySQL.Spreadsheet.FormulaEditor.validate(spreadsheet, $cell, this);
                                }
                                var column = spreadsheet.colOf($cell),
                                    length = column.length,
                                    type = column.type,
                                    $input = $cell.find("input, textarea"),
                                    cellVal = $input.size() > 0 ? $input.val() : $cell.find("div[contenteditable]").text(),
                                    valid = true,
                                    message = "";
                                if (cellVal.length > 0) {
                                    if (cellVal.charAt(0) != "=") {
                                        // Not relevant anymore
                                        $cell.removeClass("anti-formula");
                                    } else {
                                        // The cell starts with '='
                                        if (transformIfFormula === true
                                            && !$cell.hasClass("anti-formula")
                                            && spreadsheet.perms.hasPerm('EDIT_ROWS_METADATA')) {
                                            /* TODO Formula activation/deactivation here */
                                            Confluence.PlaySQL.Spreadsheet.FormulaEditor.edit(spreadsheet, $cell, this, cellVal);
                                            this.focus(true);
                                            return true;
                                        }
                                    }
                                } else {
                                    $cell.removeClass("anti-formula");
                                }
                                if (length) {
                                    if (cellVal != undefined && cellVal.length > length) {
                                        valid = false;
                                        message = "The maximum size is " + length;
                                    }
                                }
                                if (type == "INTEGER" || type == "BIG_INTEGER") {
                                    valid = valid && /^[ +-]?\d* *$/.test(cellVal);
                                    message = "An integer is expected";
                                }
                                if (valid) {
                                    $cell.removeClass("invalid").removeAttr("title");
                                    $cell.tipsy("disable");
                                    $cell.tipsy("hide");
                                } else {
                                    $cell.addClass("invalid").attr("title", message);
                                    $cell.tipsy({delayIn: 0});
                                    $cell.tipsy("enable");
                                    $cell.tipsy("show");
                                }
                                return valid;
                            },
                            /**
                             * Renders the cell with a temporary placeholder representing the value entered
                             * by the user. Then call spreadsheet.saveCell($cell, val) to actually save
                             * the cell. saveCell() is in charge of calling .view() with the server-side
                             * value.
                             *
                             * Adds the class 'saving' and removes 'editing'.
                             */
                            saveAndView: function() {
                                if (!$cell.hasClass("editing") || $cell.hasClass("saving")) return;
                                if (!this.validate()) {
                                    return;
                                }
                                var rowId = $cell.closest("tr").attr("data-row-id");
                                if (rowId == undefined || rowId == "") {
                                    return;
                                }
                                var column = spreadsheet.colOf($cell);
                                if (!spreadsheet.perms.hasPerm('EDIT_CELLS') && !spreadsheet.perms.hasPerm('Q_VIEW_EDIT', column)) return;

                                if ($cell.hasClass("formula")) {
                                    Confluence.PlaySQL.Spreadsheet.FormulaEditor.saveAndView(spreadsheet, $cell, this);
                                    return;
                                }

                                var $input = $cell.find("input, textarea"),
                                    val = $input.size() > 0 ? $input.val() : $cell.find("div[contenteditable]").text(),
                                    previousValue = $cell.attr('data-value'),
                                    rendered = this.parent.renderCellContents(val, spreadsheet, column);
                                if (val == '') {
                                    val = null;
                                }
                                if (previousValue == val) {
                                    $cell.html($cell.attr('data-rendered')).removeClass('editing');
                                    return;
                                }

                                $cell.attr('data-value', val);
                                // Keep the value for cancel
                                $cell.attr('data-rendered', rendered);

                                $cell.html(rendered).addClass("saving").removeClass("editing");
                                spreadsheet.saveCell($cell, val);
                                Confluence.PlaySQL.Spreadsheet.FormulaEditor.checkAndSaveFormulaRemoval(spreadsheet, $cell);
                            },
                            cancelEdit: function() {
                                if ($cell.hasClass("formula")) {
                                    Confluence.PlaySQL.Spreadsheet.FormulaEditor.cancelEdit(spreadsheet, $cell, this);
                                    return;
                                }
                                var rendered = $cell.attr('data-rendered');
                                if (rendered == undefined) {
                                    $cell.empty().removeClass("editing");
                                } else {
                                    $cell.html(rendered).removeClass("editing");
                                }
                            },
                            markInError: function(reason) {
                                $cell.addClass("invalid").attr("title", reason || "The cell couldn't be saved. Check the input and try again.");
                                // Transform as a tooltip
                                $cell.tooltip({delayIn: 0}); // Call Confluence's default values
                                $cell.tipsy("enable");
                                $cell.tipsy("show");
                                if (!$cell.hasClass("editing")) {
                                    this.edit();
                                    $cell.removeClass("saving");
                                }
                            }
                        };
                    return operations;
                },
                appendRendererDetailsTo: function(spreadsheet, $div, column) {
                    var formats = column.formats,
                        jointIdColumn = formats["joint-id"],
                        jointTable = formats["joint-table"],
                        jointColumn = formats["joint-column"],
                        jointFK = formats["joint-fk"];
                    if (jointColumn == column.name) jointColumn = undefined;
                    if (jointIdColumn) {
                        var idColumn = spreadsheet.colOf(jointIdColumn);
                        if (idColumn) {
                            jointTable = idColumn.formats["joint-table"];
                        }
                    }
                    if (jointIdColumn || jointTable) {
                        $div.append(Confluence.Templates.PlaySQL.columnPropertiesJoint({
                            jointIdColumn: jointIdColumn,
                            jointTable: jointTable,
                            jointColumn: jointColumn,
                            jointFK: jointFK
                        }));
                    }
                },
                getFormatValues : function($form, column) {
                    var result = {},
                        jointId = $form.find("#column-joint-id").val(),
                        jointColumn = $form.find("#column-joint-column").val(),
                        jointTable = $form.find("#column-joint-table").val(),
                        jointFK = $form.find("#column-joint-fk").val();
                    if (jointId) result["joint-id"] = jointId;
                    if (jointColumn && jointColumn != column.name) result["joint-column"] = jointColumn;
                    if (jointTable) result["joint-table"] = jointTable;
                    if (jointFK) result["joint-fk"] = jointFK;
                    return result;
                }
            };
        };

        Confluence.PlaySQL.Spreadsheet.RendererBase = RendererBase;
        Confluence.PlaySQL.Spreadsheet.rendererBase = new RendererBase();

        var DefaultRenderer = function() {
            var base = Confluence.PlaySQL.Spreadsheet.rendererBase;
            return $.extend({}, base, {
                key: "as-string",
                getI18nName : function() {
                    return AJS.I18n.getText("com.playsql.playsql-base-plugin.renderers.as-string")
                },
                isSupportedType : function(type) {
                    return "VARCHAR" == type || "TEXT" == type;
                },
                getDefaultType : function() {
                    return "TEXT";
                },
                isSystemRenderer : function() {
                    return false;
                },
                createInputBox : function(value) {
                    return this.createTextareaBox(value);
                }
            });
        };
        api.register(new DefaultRenderer());

        function checkNotNull(value, message) {
            if (value == undefined) {
                throw "Unexpectedly undefined value: " + value + "(" + message + ")";
            }
            if (value instanceof $) {
                if (value.size() == 0)
                    throw "Unexpectedly empty jquery element: " + value + "(" + message + ")";
            }
            return value;
        }
    });

})(AJS.$);